__title__ = "NSoL"
__author__ = "Michael Ebner"
__email__ = "michael.ebner.14@ucl.ac.uk"
__license__ = "BSD-3-Clause"
__summary__ = "The Numerical Solver Library (NSoL) is a research-focused " \
    "toolkit for denoising and deconvolution applications using ADMM and " \
    "Primal-Dual methods."
__uri__ = "https://github.com/gift-surg/NSoL"
__version__ = "0.1.6.2"
