from nsol.__about__ import (
    __author__,
    __email__,
    __license__,
    __summary__,
    __title__,
    __uri__,
    __version__,
 )

__all__ = [
    "__author__",
    "__email__",
    "__license__",
    "__summary__",
    "__title__",
    "__uri__",
    "__version__",
]
